
# THIS FILE IS GENERATED FROM TINT SETUP.PY
short_version = '0.1.0'
version = '0.1.0'
full_version = '0.1.0.dev+d933665'
git_revision = 'd933665f514dfd429cf3691f87d500b370290873'
release = False

if not release:
    version = full_version



import copy
import datetime
import sys
import numpy as np
import pandas as pd

from .grid_utils import get_grid_size, get_radar_info, extract_grid_data
from .helpers import Record, Counter
from .phase_correlation import get_global_shift
from .matching import get_pairs
from .objects import init_current_objects, update_current_objects
from .objects import get_object_prop, write_tracks

# Tracking Parameter Defaults
FIELD_THRESH = 32
ISO_THRESH = 4
ISO_SMOOTH = 4
MIN_SIZE = 8
SEARCH_MARGIN = 250
FLOW_MARGIN = 750
MAX_DISPARITY = 999
MAX_FLOW_MAG = 50
MAX_SHIFT_DISP = 15
GS_ALT = 1500
"""
Tracking Parameter Guide
------------------------
FIELD_THRESH : units of 'field' attribute
    The threshold used for object detection. Detected objects are connnected
    pixels above this threshold.
ISO_THRESH : units of 'field' attribute
    Used in isolated cell classification. Isolated cells must not be connected
    to any other cell by contiguous pixels above this threshold.
ISO_SMOOTH : pixels
    Gaussian smoothing parameter in peak detection preprocessing. See
    single_max in tint.objects.
MIN_SIZE : square kilometers
    The minimum size threshold in pixels for an object to be detected.
SEARCH_MARGIN : meters
    The radius of the search box around the predicted object center.
FLOW_MARGIN : meters
    The margin size around the object extent on which to perform phase
    correlation.
MAX_DISPARITY : float
    Maximum allowable disparity value. Larger disparity values are sent to
    LARGE_NUM.
MAX_FLOW_MAG : meters per second
    Maximum allowable global shift magnitude. See get_global_shift in
    tint.phase_correlation.
MAX_SHIFT_DISP : meters per second
    Maximum magnitude of difference in meters per second for two shifts to be
    considered in agreement. See correct_shift in tint.matching.
GS_ALT : meters
    Altitude in meters at which to perform phase correlation for global shift
    calculation. See correct_shift in tint.matching.
"""


class Cell_tracks(object):
    def __init__(self, field='reflectivity'):
        self.params = {'FIELD_THRESH': FIELD_THRESH,
                       'MIN_SIZE': MIN_SIZE,
                       'SEARCH_MARGIN': SEARCH_MARGIN,
                       'FLOW_MARGIN': FLOW_MARGIN,
                       'MAX_FLOW_MAG': MAX_FLOW_MAG,
                       'MAX_DISPARITY': MAX_DISPARITY,
                       'MAX_SHIFT_DISP': MAX_SHIFT_DISP,
                       'ISO_THRESH': ISO_THRESH,
                       'ISO_SMOOTH': ISO_SMOOTH,
                       'GS_ALT': GS_ALT}

        self.field = field
        self.grid_size = None
        self.radar_info = None
        self.last_grid = None
        self.counter = None
        self.record = None
        self.current_objects = None
        self.tracks = pd.DataFrame()
        self.odong = None
        self.__saved_record = None
        self.__saved_counter = None
        self.__saved_objects = None

    def __save(self):
        """ Saves deep copies of record, counter, and current_objects. """
        self.__saved_record = copy.deepcopy(self.record)
        self.__saved_counter = copy.deepcopy(self.counter)
        self.__saved_objects = copy.deepcopy(self.current_objects)

    def __load(self):
        """ Loads saved copies of record, counter, and current_objects. If new
        tracks are appended to existing tracks via the get_tracks method, the
        most recent scan prior to the addition must be overwritten to link up
        with the new scans. Because of this, record, counter and
        current_objects must be reverted to their state in the penultimate
        iteration of the loop in get_tracks. See get_tracks for details. """
        self.record = self.__saved_record
        self.counter = self.__saved_counter
        self.current_objects = self.__saved_objects

    def get_tracks(self, grids, c):
        """ Obtains tracks given a list of data arrays. This is the
        primary method of the tracks class. This method makes use of all of the
        functions and helper classes defined above. """
        start_time = datetime.datetime.now()
        ncells = 0
        if self.record is None:
            # tracks object being initialized
            grid_obj2 = next(grids)
            self.grid_size = get_grid_size(grid_obj2)
            self.radar_info = get_radar_info(c)
            self.counter = Counter()
            self.record = Record(grid_obj2)
        else:
            # tracks object being updated
            grid_obj2 = self.last_grid
            self.tracks.drop(self.record.scan + 1)  # last scan is overwritten

        if self.current_objects is None:
            newRain = True
        else:
            newRain = False
        raw2, frame2 = extract_grid_data(grid_obj2, self.field, self.grid_size,
                                         self.params)
        self.odong=[]
        while grid_obj2 is not None:
            grid_obj1 = grid_obj2
            raw1 = raw2
            frame1 = frame2

            try:
                grid_obj2 = next(grids)
            except StopIteration:
                grid_obj2 = None

            if grid_obj2 is not None:
                self.record.update_scan_and_time(grid_obj1, grid_obj2)
                raw2, frame2 = extract_grid_data(grid_obj2,
                                                 self.field,
                                                 self.grid_size,
                                                 self.params)
            else:
                self.__save()
                self.last_grid = grid_obj1
                self.record.update_scan_and_time(grid_obj1)
                raw2 = None
                frame2 = np.zeros_like(frame1)

            if np.nanmax(frame1) == 0:
                newRain = True

                sys.stdout.flush()
                sys.stdout.write('\rNo cells found in scan %s     '%(self.record.scan))
                sys.stdout.flush()
                self.current_objects = None
                continue
            ncells += 1
            global_shift = get_global_shift(raw1, raw2, self.params)		
            pairs = get_pairs(frame1,
                              frame2,
                              global_shift,
                              self.current_objects,
                              self.record,
                              self.params)
            if newRain:
                # first nonempty scan after a period of empty scans
                self.current_objects, self.counter = init_current_objects(
                    frame1,
                    frame2,
                    pairs,
                    self.counter
                )
                newRain = False
            else:
                self.current_objects, self.counter = update_current_objects(
                    frame1,
                    frame2,
                    pairs,
                    self.current_objects,
                    self.counter
                )
            #try:
            obj_props = get_object_prop(frame1, grid_obj1, self.field,
                                            self.record, self.params)
            
            self.record.add_uids(self.current_objects)
            self.tracks = write_tracks(self.tracks, self.record,
                                       self.current_objects, obj_props)
            #except IndexError:
            #obj_props = None
            self.odong.append(frame1)
            del grid_obj1, raw1, frame1, global_shift, pairs, obj_props
            # scan loop end
        self.__load()
        time_elapsed = datetime.datetime.now() - start_time
        print('time elapsed', np.round(time_elapsed.seconds/60, 1), 'minutes')
        return ncells



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import ndimage
import sys
from .grid_utils import get_filtered_frame
from scipy.spatial import ConvexHull, convex_hull_plot_2d

def get_object_center(obj_id, labeled_image):
    """ Returns index of center pixel of the given object id from labeled
    image. The center is calculated as the median pixel of the object extent;
    it is not a true centroid. """
    obj_index = np.argwhere(labeled_image == obj_id)
    center = np.median(obj_index, axis=0).astype('i')
    return center


def get_obj_extent(labeled_image, obj_label):
    """ Takes in labeled image and finds the radius, area, and center of the
    given object. """
    obj_index = np.argwhere(labeled_image == obj_label)#vi tri lab

    xlength = np.max(obj_index[:, 0]) - np.min(obj_index[:, 0]) + 1
    ylength = np.max(obj_index[:, 1]) - np.min(obj_index[:, 1]) + 1
    obj_radius = np.max((xlength, ylength))/2#ban kinh max
    obj_center = np.round(np.median(obj_index, axis=0), 0)#vitri trung tam
    obj_area = len(obj_index[:, 0])#pham vi o dong

    obj_extent = {'obj_center': obj_center, 'obj_radius': obj_radius,
                  'obj_area': obj_area, 'obj_index': obj_index}
    return obj_extent


def init_current_objects(first_frame, second_frame, pairs, counter):
    """ Returns a dictionary for objects with unique ids and their
    corresponding ids in frame1 and frame1. This function is called when
    echoes are detected after a period of no echoes. """
    nobj = np.max(first_frame)

    id1 = np.arange(nobj) + 1
    uid = counter.next_uid(count=nobj)
    id2 = pairs
    obs_num = np.zeros(nobj, dtype='i')
    origin = np.array(['-1']*nobj)

    current_objects = {'id1': id1, 'uid': uid, 'id2': id2,
                       'obs_num': obs_num, 'origin': origin}
    current_objects = attach_last_heads(first_frame, second_frame,
                                        current_objects)
    return current_objects, counter


def update_current_objects(frame1, frame2, pairs, old_objects, counter):
    """ Removes dead objects, updates living objects, and assigns new uids to
    new-born objects. """
    nobj = np.max(frame1)
    id1 = np.arange(nobj) + 1
    uid = np.array([], dtype='str')
    obs_num = np.array([], dtype='i')
    origin = np.array([], dtype='str')

    for obj in np.arange(nobj) + 1:
        if obj in old_objects['id2']:
            obj_index = old_objects['id2'] == obj
            uid = np.append(uid, old_objects['uid'][obj_index])
            obs_num = np.append(obs_num, old_objects['obs_num'][obj_index] + 1)
            origin = np.append(origin, old_objects['origin'][obj_index])
        else:
            #  obj_orig = get_origin_uid(obj, frame1, old_objects)
            obj_orig = '-1'
            origin = np.append(origin, obj_orig)
            if obj_orig != '-1':
                uid = np.append(uid, counter.next_cid(obj_orig))
            else:
                uid = np.append(uid, counter.next_uid())
            obs_num = np.append(obs_num, 0)

    id2 = pairs
    current_objects = {'id1': id1, 'uid': uid, 'id2': id2,
                       'obs_num': obs_num, 'origin': origin}
    current_objects = attach_last_heads(frame1, frame2, current_objects)
    return current_objects, counter


def attach_last_heads(frame1, frame2, current_objects):
    """ Attaches last heading information to current_objects dictionary. """
    nobj = len(current_objects['uid'])
    heads = np.ma.empty((nobj, 2))
    for obj in range(nobj):
        if ((current_objects['id1'][obj] > 0) and
                (current_objects['id2'][obj] > 0)):
            center1 = get_object_center(current_objects['id1'][obj], frame1)
            center2 = get_object_center(current_objects['id2'][obj], frame2)
            heads[obj, :] = center2 - center1
        else:
            heads[obj, :] = np.ma.array([-999, -999], mask=[True, True])

    current_objects['last_heads'] = heads
    return current_objects


def check_isolation(raw, filtered, grid_size, params):
    """ Returns list of booleans indicating object isolation. Isolated objects
    are not connected to any other objects by pixels greater than ISO_THRESH,
    and have at most one peak. """
    nobj = np.max(filtered)
    min_size = params['MIN_SIZE'] / np.prod(grid_size[1:]/1000)
    iso_filtered = get_filtered_frame(raw,
                                      min_size,
                                      params['ISO_THRESH'])
    nobj_iso = np.max(iso_filtered)
    iso = np.empty(nobj, dtype='bool')

    for iso_id in np.arange(nobj_iso) + 1:
        obj_ind = np.where(iso_filtered == iso_id)
        objects = np.unique(filtered[obj_ind])
        objects = objects[objects != 0]
        if len(objects) == 1 and single_max(obj_ind, raw, params):
            iso[objects - 1] = True
        else:
            iso[objects - 1] = False
    return iso


def single_max(obj_ind, raw, params):
    """ Returns True if object has at most one peak. """
    max_proj = np.max(raw, axis=0)
    smooth = ndimage.filters.gaussian_filter(max_proj, params['ISO_SMOOTH'])
    padded = np.pad(smooth, 1, mode='constant')
    obj_ind = [axis + 1 for axis in obj_ind]  # adjust for padding
    maxima = 0
    for pixel in range(len(obj_ind[0])):
        ind_0 = obj_ind[0][pixel]
        ind_1 = obj_ind[1][pixel]
        neighborhood = padded[(ind_0-1):(ind_0+2), (ind_1-1):(ind_1+2)]
        max_ind = np.unravel_index(neighborhood.argmax(), neighborhood.shape)
        if max_ind == (1, 1):
            maxima += 1
            if maxima > 1:
                return False
    return True


def get_object_prop(image1, grid1, field, record, params):
    """ Returns dictionary of object properties for all objects found in
    image1. """
    id1 = []
    center = []
    grid_x = []
    grid_y = []
    area = []
    longitude = []
    latitude = []
    field_max = []
    field_mean = []
    nobj = np.max(image1)
    #soodong=[]
    unit_dim = record.grid_size
    unit_alt = unit_dim[0]/1000
    unit_area = 1 #(unit_dim[1]*unit_dim[2])/(1000**2)
    unit_vol = (unit_dim[0]*unit_dim[1]*unit_dim[2])/(1000**3)
    raw3D = grid1['data']
    get_items = []
    lonoj=[]
    latoj=[]
    for obj in np.arange(nobj) + 1:
        try:

            obj_index = np.argwhere(image1 == obj)
            this_centroid = np.round(np.mean(obj_index, axis=0), 3)
            rounded = np.round(this_centroid).astype('i')
            c_x = grid1['x'][rounded[1]]
            c_y = grid1['y'][rounded[0]]
            longitude.append(np.round(grid1['x'][rounded[1]], 4))
            latitude.append(np.round(grid1['y'][rounded[0]], 4))
            


            id1.append(obj)

            # 2D frame stats
            center.append(np.median(obj_index, axis=0))
            grid_x.append(this_centroid[1])
            grid_y.append(this_centroid[0])
            area.append(obj_index.shape[0] * unit_area)
            lo_A=grid1['x'][obj_index[:,1]]
            la_A=grid1['y'][obj_index[:,0]]	
			
            points = np.vstack((lo_A,la_A)).T
            hull = ConvexHull(points)
            lonoj.append(points[hull.simplices][:,:,0].ravel().astype('float32'))
            latoj.append(points[hull.simplices][:,:,1].ravel().astype('float32'))			
            # raw 3D grid stats
            obj_slices = [raw3D[:, ind[0], ind[1]] for ind in obj_index]
            field_max.append(np.nanmax(obj_slices))
            field_mean.append(np.nanmean(obj_slices))
            filtered_slices = [obj_slice > params['FIELD_THRESH']
                               for obj_slice in obj_slices]
            get_items.append(obj - 1)
        except IndexError:
            pass
    # cell isolation
    isolation = check_isolation(raw3D, image1, record.grid_size, params)

    objprop = {'id1': id1,
               'center': center,
               'grid_x': grid_x,
               'grid_y': grid_y,
               'area': area,
               'field_max': field_max,
               'field_mean':field_mean,
               'lon': longitude,
               'lat': latitude,
               'isolated': isolation,
               'ok_items': get_items,
               'lonoj': lonoj,
               'latoj': latoj}
    return objprop


def write_tracks(old_tracks, record, current_objects, obj_props):
    """ Writes all cell information to tracks dataframe. """
    sys.stdout.flush()
    sys.stdout.write('\rWriting tracks for scan %s \n'%(record.scan))
    sys.stdout.flush()

    nobj = len(obj_props['id1'])
    scan_num = [record.scan] * nobj
    gi = obj_props['ok_items']
    uid = current_objects['uid'][gi]
    new_tracks = pd.DataFrame({
        'scan': scan_num,
        'uid': uid,
        'time': record.time,
        'grid_x': obj_props['grid_x'],
        'grid_y': obj_props['grid_y'],
        'lon': obj_props['lon'],
        'lat': obj_props['lat'],
        'area': obj_props['area'],
        'max': obj_props['field_max'],
        'mean': obj_props['field_mean'],
        'isolated': obj_props['isolated'][gi],
        'lonoj': obj_props['lonoj'],
        'latoj': obj_props['latoj']		
    })
    new_tracks.set_index(['scan', 'uid'], inplace=True)
    tracks = old_tracks.append(new_tracks)
    return tracks



import numpy as np
from scipy import optimize
from .phase_correlation import get_ambient_flow
from .objects import get_obj_extent
LARGE_NUM = 1000
def euclidean_dist(vec1, vec2):
    """ Computes euclidean distance. """
    vec1 = np.array(vec1)
    vec2 = np.array(vec2)
    dist = np.sqrt(sum((vec1-vec2)**2))
    return dist


def get_sizeChange(size1, size2):
    """ Returns change in size of an echo as the ratio of the larger size to
    the smaller, minus 1. """
    if (size1 < 5) and (size2 < 5):
        return 0
    elif size1 >= size2:
        return size1/size2 - 1
    else:
        return size2/size1 - 1


def find_objects(search_box, image2):
    """ Identifies objects found in the search region. """
    if not search_box['valid']:
        obj_found = np.array(-1)
    else:
        search_area = image2[search_box['x1']:search_box['x2'],
                             search_box['y1']:search_box['y2']]
        obj_found = np.unique(search_area)
    return obj_found


def shifts_disagree(shift1, shift2, record, params):
    """ Returns True if shift disparity greater than MAX_SHIFT_DISP
    parameter. """
    shift1 = shift1*record.grid_size[1:]
    shift2 = shift2*record.grid_size[1:]
    shift_disparity = euclidean_dist(shift1, shift2)
    return shift_disparity/record.interval.seconds > params['MAX_SHIFT_DISP']


def clip_shift(shift, record, params):
    """ Clips shift according to MAX_FLOW_MAG paramter. """#so sanh vs toc do toi da m/s
    shift_meters = shift * record.grid_size[1:]
    shift_mag = np.linalg.norm(shift_meters)
    velocity = shift_mag/record.interval.seconds
    unit = shift_meters/shift_mag
    #print(velocity,unit,record.grid_size[1:],record.interval.seconds)
    if velocity > params['MAX_FLOW_MAG']:
        clipped = unit * params['MAX_FLOW_MAG'] * record.interval.seconds
        clipped_pix = clipped/record.grid_size[1:]
        return clipped_pix
    else:
        return shift


def correct_shift(local_shift, current_objects, obj_id1, global_shift, record,
                  params):
    """ Takes in flow vector based on local phase correlation (see
    get_std_flow) and compares it to the last headings of the object and
    the global_shift vector for that timestep. Corrects accordingly.
    Note: At the time of this function call, current_objects has not yet been
    updated for the current frame1 and frame2, so the id2s in current_objects
    correspond to the objects in the current frame1. """
    global_shift = clip_shift(global_shift, record, params)
    #print('global_shift in correct_shift',global_shift)
    if current_objects is None:
        last_heads = None
    else:
        obj_index = current_objects['id2'] == obj_id1
        last_heads = current_objects['last_heads'][obj_index].flatten()
        last_heads = np.round(last_heads * record.interval_ratio, 2)
        if len(last_heads) == 0:
            last_heads = None

    if last_heads is None:
        if shifts_disagree(local_shift, global_shift, record, params):
            case = 0
            corrected_shift = global_shift
        else:
            case = 1
            corrected_shift = (local_shift + global_shift)/2

    elif shifts_disagree(local_shift, last_heads, record, params):
        if shifts_disagree(local_shift, global_shift, record, params):
            case = 2
            corrected_shift = last_heads
        else:
            case = 3
            corrected_shift = local_shift

    else:
        case = 4
        corrected_shift = (local_shift + last_heads)/2

    corrected_shift = np.round(corrected_shift, 2)

    record.count_case(case)
    record.record_shift(corrected_shift, global_shift,
                        last_heads, local_shift, case)
    return corrected_shift#hieu chinh lai vectoc


def predict_search_extent(obj1_extent, shift, params, grid_size):
    """ Predicts search extent/region for the object in image2 given
    the image shift. """# tim vung anh huong cua o dong du doan
    shifted_center = obj1_extent['obj_center'] + shift
    search_radius_r = params['SEARCH_MARGIN'] / grid_size[1]
    search_radius_c = params['SEARCH_MARGIN'] / grid_size[2]
    x1 = shifted_center[0] - search_radius_r
    x2 = shifted_center[0] + search_radius_r + 1
    y1 = shifted_center[1] - search_radius_c
    y2 = shifted_center[1] + search_radius_c + 1
    x1 = np.int(x1)
    x2 = np.int(x2)
    y1 = np.int(y1)
    y2 = np.int(y2)
    return {'x1': x1, 'x2': x2, 'y1': y1, 'y2': y2,
            'center_pred': shifted_center, 'valid': True}


def check_search_box(search_box, img_dims):
    """ Checks if search_box is within the boundaries of the frame. Clips to
    edges of frame if out of bounds. Marks as invalid if too small. """
    if search_box['x1'] < 0:
        search_box['x1'] = 0
    if search_box['y1'] < 0:
        search_box['y1'] = 0
    if search_box['x2'] > img_dims[0]:
        search_box['x2'] = img_dims[0]
    if search_box['y2'] > img_dims[1]:
        search_box['y2'] = img_dims[1]
    if ((search_box['x2'] - search_box['x1'] < 5) or
            (search_box['y2'] - search_box['y1'] < 5)):
        search_box['valid'] = False
    return search_box


def get_disparity(obj_found, image2, search_box, obj1_extent):
    """ Computes disparities for objects in obj_found. """
    dist_pred = np.empty(0)
    change = np.empty(0)
    for target_obj in obj_found:
        target_extent = get_obj_extent(image2, target_obj)
        euc_dist = euclidean_dist(target_extent['obj_center'],
                                  search_box['center_pred'])
        dist_pred = np.append(dist_pred, euc_dist)
        size_changed = get_sizeChange(target_extent['obj_area'],
                                      obj1_extent['obj_area'])
        change = np.append(change, size_changed)

    disparity = dist_pred + change
    return disparity


def get_disparity_all(obj_found, image2, search_box, obj1_extent):
    """ Returns disparities of all objects found within the search box. """
    if np.max(obj_found) <= 0:
        disparity = np.array([LARGE_NUM])
    else:
        obj_found = obj_found[obj_found > 0]
        disparity = get_disparity(obj_found, image2,
                                  search_box, obj1_extent)
    return disparity


def save_obj_match(obj_id1, obj_found, disparity, obj_match, params):
    """ Saves disparity values in obj_match matrix. If disparity is greater than
    MAX_DISPARITY, saves a large number. """
    disparity[disparity > params['MAX_DISPARITY']] = LARGE_NUM
    if np.max(obj_found) > 0:
        obj_found = obj_found[obj_found > 0]
        obj_found = obj_found - 1
        obj_id1 = obj_id1 - 1
        obj_match[obj_id1, obj_found] = disparity
    return obj_match


def locate_allObjects(image1, image2, global_shift, current_objects, record,
                      params):
    """ Matches all the objects in image1 to objects in image2. This is the
    main function called on a pair of images. """#ket noi id
    nobj1 = np.max(image1)
    nobj2 = np.max(image2)

    if (nobj2 == 0) or (nobj1 == 0):
        print('No echoes to track!')
        return

    obj_match = np.full((nobj1, np.max((nobj1, nobj2))),
                        LARGE_NUM, dtype='f')

    index_obj={}
    for obj_id1 in np.arange(nobj1) + 1:
        obj1_extent = get_obj_extent(image1, obj_id1)#dua ra cac ttobj_extent = {'obj_center': obj_center, 'obj_radius': obj_radius,
																				#'obj_area': obj_area, 'obj_index': obj_index}
        #print('obj1_extent',obj1_extent['obj_index'])	
        index_obj[str(obj_id1)]=	obj1_extent['obj_index']#them 27042021
        shift = get_ambient_flow(obj1_extent, image1,
                                 image2, params, record.grid_size)#lay vectoc cua 1 o dong obj_id1
        if shift is None:
            record.count_case(5)
            shift = global_shift

        shift = correct_shift(shift, current_objects, obj_id1,
                              global_shift, record, params)

        search_box = predict_search_extent(obj1_extent, shift,
                                           params, record.grid_size)# du doan khung anh huong của o dong
        search_box = check_search_box(search_box, image2.shape)# check pham vi co nam trong khung 2
        objs_found = find_objects(search_box, image2)# tim tat ca cac o dong tren khung thuc te o thoi diem sau.
        disparity = get_disparity_all(objs_found, image2,
                                      search_box, obj1_extent)# tinh toan khoang cach cua odong k1 vs ca odong tim dc trong k2
        #print('disparity',disparity)						  
        obj_match = save_obj_match(obj_id1, objs_found, disparity, obj_match,
                                   params)							   
								   
    return obj_match,index_obj


def match_pairs(obj_match, params):
    """ Matches objects into pairs given a disparity matrix and removes
    bad matches. Bad matches have a disparity greater than the maximum
    threshold. """
    pairs = optimize.linear_sum_assignment(obj_match)

    for id1 in pairs[0]:
        if obj_match[id1, pairs[1][id1]] > params['MAX_DISPARITY']:
            pairs[1][id1] = -1  # -1 indicates the object has died

    pairs = pairs[1] + 1  # ids in current_objects are 1-indexed
    return pairs


def get_pairs(image1, image2, global_shift, current_objects, record, params):
    """ Given two images, this function identifies the matching objects and
    pairs them appropriately. See disparity function. """ #image1 -lab,global_shift - vectoc tong,
    nobj1 = np.max(image1)
    nobj2 = np.max(image2)

    if nobj1 == 0:
        print('No echoes found in the first scan.')
        return
    elif nobj2 == 0:
        zero_pairs = np.zeros(nobj1)
        return zero_pairs

    obj_match,index_obj = locate_allObjects(image1,
                                  image2,
                                  global_shift,
                                  current_objects,
                                  record,
                                  params)
    pairs = match_pairs(obj_match, params)
    return pairs



import datetime
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import ndimage

def parse_grid_datetime(grid_obj):
    """ Obtains datetime object from the data dictionary. """
    #dt_string = grid_obj.time['units'].split(' ')[-1]
    #date = dt_string[:10]
    ##time = dt_string[11:19]
    #dt = datetime.datetime.strptime(date + ' ' + time, '%Y-%m-%d %H:%M:%S')
    return grid_obj['time']


def get_grid_size(grid_obj):
    """ Calculates grid size per dimension given a grid object. """
    return np.array(grid_obj['data'].shape)


def get_radar_info(radar):
    info = {'radar_lon': radar[0],
            'radar_lat': radar[0]}
    return info


def get_grid_alt(grid_size, alt_meters=1500):
    """ Returns z-index closest to alt_meters. """
    return np.int(np.round(alt_meters/grid_size[0]))


def get_vert_projection(grid, thresh=40):
    """ Returns boolean vertical projection from grid. """
    return np.any(grid > thresh, axis=0)


def get_filtered_frame(grid, min_size, thresh):
    """ Returns a labeled frame from gridded radar data. Smaller objects
    are removed and the rest are labeled. """
    echo_height = get_vert_projection(grid, thresh)
    labeled_echo = ndimage.label(echo_height)[0]
    frame = clear_small_echoes(labeled_echo, min_size)
    return frame


def clear_small_echoes(label_image, min_size):
    """ Takes in binary image and clears objects less than min_size. """
    flat_image = pd.Series(label_image.flatten())
    flat_image = flat_image[flat_image > 0]
    size_table = flat_image.value_counts(sort=False)
    small_objects = size_table.keys()[size_table < min_size]

    for obj in small_objects:
        label_image[label_image == obj] = 0
    label_image = ndimage.label(label_image)
    return label_image[0]


def extract_grid_data(grid_obj, field, grid_size, params):
    """ Returns filtered grid frame and raw grid slice at global shift
    altitude. """
    try:
        masked = grid_obj['data'].filled(0)
    except AttributeError:
        masked = grid_obj['data']
    gs_alt = params['GS_ALT']

    raw = masked[0 , :, :]#value
    frame = get_filtered_frame(masked, params['MIN_SIZE'],
                               params['FIELD_THRESH'])
    return raw, frame



import numpy as np
from scipy import ndimage
def get_ambient_flow(obj_extent, img1, img2, params, grid_size):
    margin_r = params['FLOW_MARGIN'] / grid_size[1]
    margin_c = params['FLOW_MARGIN'] / grid_size[2]
    row_lb = obj_extent['obj_center'][0] - obj_extent['obj_radius'] - margin_r
    row_ub = obj_extent['obj_center'][0] + obj_extent['obj_radius'] + margin_r
    col_lb = obj_extent['obj_center'][1] - obj_extent['obj_radius'] - margin_c
    col_ub = obj_extent['obj_center'][1] + obj_extent['obj_radius'] + margin_c
    row_lb = np.int(row_lb)
    row_ub = np.int(row_ub)
    col_lb = np.int(col_lb)
    col_ub = np.int(col_ub)

    dims = img1.shape

    row_lb = np.max([row_lb, 0])
    row_ub = np.min([row_ub, dims[0]])
    col_lb = np.max([col_lb, 0])
    col_ub = np.max([col_ub, dims[1]])

    flow_region1 = np.copy(img1[row_lb:row_ub+1, col_lb:col_ub+1])
    flow_region2 = np.copy(img2[row_lb:row_ub+1, col_lb:col_ub+1])

    flow_region1[flow_region1 != 0] = 1
    flow_region2[flow_region2 != 0] = 1#mang 2 chieu
    return fft_flowvectors(flow_region1, flow_region2)


def fft_flowvectors(im1, im2, global_shift=False):
    """ Estimates flow vectors in two images using cross covariance. """
    if not global_shift and (np.max(im1) == 0 or np.max(im2) == 0):
        return None

    crosscov = fft_crosscov(im1, im2)
    sigma = (1/8) * min(crosscov.shape)
    cov_smooth = ndimage.filters.gaussian_filter(crosscov, sigma)
    dims = np.array(im1.shape)
    pshift = np.argwhere(cov_smooth == np.max(cov_smooth))[0]
    pshift = (pshift+1) - np.round(dims/2, 0)
    return pshift


def fft_crosscov(im1, im2):
    """ Computes cross correlation matrix using FFT method. """
    fft1_conj = np.conj(np.fft.fft2(im1))
    fft2 = np.fft.fft2(im2)
    normalize = abs(fft2*fft1_conj)
    normalize[normalize == 0] = 1  
    cross_power_spectrum = (fft2*fft1_conj)/normalize
    crosscov = np.fft.ifft2(cross_power_spectrum)
    crosscov = np.real(crosscov)
    return fft_shift(crosscov)


def fft_shift(fft_mat):
    """ Rearranges the cross correlation matrix so that 'zero' frequency or DC
    component is in the middle of the matrix. Taken from stackoverflow Que.
    30630632. """
    if type(fft_mat) is np.ndarray:
        rd2 = np.int(fft_mat.shape[0]/2)
        cd2 = np.int(fft_mat.shape[1]/2)
        quad1 = fft_mat[:rd2, :cd2]
        quad2 = fft_mat[:rd2, cd2:]
        quad3 = fft_mat[rd2:, cd2:]
        quad4 = fft_mat[rd2:, :cd2]
        centered_t = np.concatenate((quad4, quad1), axis=0)
        centered_b = np.concatenate((quad3, quad2), axis=0)
        centered = np.concatenate((centered_b, centered_t), axis=1)
        return centered
    else:
        print('input to fft_shift() should be a matrix')
        return


def get_global_shift(im1, im2, params):
    """ Returns standardazied global shift vector. im1 and im2 are full frames
    of raw DBZ values. """
    if im2 is None:
        return None
    shift = fft_flowvectors(im1, im2, global_shift=True)
    return shift



import string
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from netCDF4 import num2date, date2num
from .grid_utils import parse_grid_datetime, get_grid_size
class Counter(object):
    """
    Counter objects generate and keep track of unique cell ids.
    Currently only the uid attribute is used, but this framework can
    accomodate further development of merge/split detection.

    Attributes
    ----------
    uid : int
        Last uid assigned.
    cid : dict
        Record of cell genealogy.

    """

    def __init__(self):
        """ uid is an integer that tracks the number of independently formed
        cells. The cid dictionary keeps track of 'children' --i.e., cells that
        have split off from another cell. """
        self.uid = -1
        self.cid = {}

    def next_uid(self, count=1):
        """ Incremented for every new independently formed cell. """
        new_uids = self.uid + np.arange(count) + 1
        self.uid += count
        return np.array([str(uid) for uid in new_uids])

    def next_cid(self, pid):
        """ Returns parent uid with appended letter to denote child. """
        if pid in self.cid.keys():
            self.cid[pid] += 1
        else:
            self.cid[pid] = 0
        letter = string.ascii_lowercase[self.cid[pid]]
        return pid + letter


class Record(object):
    """
    Record objects keep track of information related to the shift correction
    process.

    Attributes
    ----------
    scan : int
        Index of the current scan.
    time : datetime
        Time corresponding to scan.
    interval : timedelta
        Temporal difference between the next scan and the current scan.
    interval_ratio : float
        Ratio of current interval to previous interval.
    grid_size : array of floats
        Length 3 array containing z, y, and x mesh size in meters.
    shifts : dataframe
        Records inputs of shift correction process. See matching.correct_shift.
    new_shfits : dataframe
        Row of new shifts to be added to shifts dataframe.
    correction_tally : dict
        Tallies correction cases for performance analysis.


    Shift Correction Case Guide:
    case0 - new object, local_shift and global_shift disagree, returns global
    case1 - new object, returns local_shift
    case2 - local disagrees with last head and global, returns last head
    case3 - local disagrees with last head, returns local
    case4 - local and last head agree, returns average of both
    case5 - flow regions empty or at edge of frame, returns global_shift

    """

    def __init__(self, grid_obj):
        self.scan = -1
        self.time = None
        self.interval = None
        self.interval_ratio = None
        self.grid_size = get_grid_size(grid_obj)
        self.shifts = pd.DataFrame()
        self.new_shifts = pd.DataFrame()
        self.correction_tally = {'case0': 0, 'case1': 0, 'case2': 0,
                                 'case3': 0, 'case4': 0, 'case5': 0}

    def count_case(self, case_num):
        """ Updates correction_tally dictionary. This is used to monitor the
        shift correction process. """
        self.correction_tally['case' + str(case_num)] += 1

    def record_shift(self, corr, gl_shift, l_heads, local_shift, case):
        """ Records corrected shift, phase shift, global shift, and last
        heads per object per timestep. This information can be used to
        monitor and refine the shift correction algorithm in the
        correct_shift function. """
        if l_heads is None:
            l_heads = np.ma.array([-999, -999], mask=[True, True])

        new_shift_record = pd.DataFrame()
        new_shift_record['scan'] = [self.scan]
        new_shift_record['uid'] = ['uid']
        new_shift_record['corrected'] = [corr]
        new_shift_record['global'] = [gl_shift]
        new_shift_record['last_heads'] = [l_heads]
        new_shift_record['phase'] = [local_shift]
        new_shift_record['case'] = [case]

        self.new_shifts = self.new_shifts.append(new_shift_record)

    def add_uids(self, current_objects):
        """ Because of the chronology of the get_tracks process, object uids
        cannot be added to the shift record at the time of correction, so they
        must be added later in the process. """
        if len(self.new_shifts) > 0:
            self.new_shifts['uid'] = current_objects['uid']
            self.new_shifts.set_index(['scan', 'uid'], inplace=True)
            self.shifts = self.shifts.append(self.new_shifts)
            self.new_shifts = pd.DataFrame()

    def update_scan_and_time(self, grid_obj1, grid_obj2=None):
        """ Updates the scan number and associated time. This information is
        used for obtaining object properties as well as for the interval ratio
        correction of last_heads vectors. """
        self.scan += 1
        self.time = parse_grid_datetime(grid_obj1)
        if grid_obj2 is None:
            # tracks for last scan are being written
            return
        time2 = parse_grid_datetime(grid_obj2)
        old_diff = self.interval
        self.interval = time2 - self.time
        if old_diff is not None:
            self.interval_ratio = self.interval.seconds/old_diff.seconds


def spl(present, time):
    out = []
    start = True
    a = ''.join(list(present.astype(int).astype(str)))
    b = list(present.astype(int).astype(str))
    ii = 0
    for k, g in groupby(a):
        gg = list(g)
        if len(gg) == 1:
            b[ii] = str(1 - int(gg[0]))
        ii += len(gg)
    ii = 0
    for k, g in groupby(''.join(b)):
        G = list(g)
        if k == '1':
            kk = 0
            for i in G:
                d1 = time[kk+ii]
                try:
                    d2 = time[kk+ii+1]
                except IndexError:
                    break
                if (d2 - d1) > 24*60**2:
                    break
                kk += 1
            if (ii+kk-1) - ii > 5:
                out.append((ii, ii+kk-1))
        ii += len(G)
    return out


def get_grids(group, slices, lon, lat, varname='rain_rate'):
    x = group.variables[varname].shape[1]
    y = group.variables[varname].shape[2]

    for s in range(slices[0], slices[-1]+1):
        yield {'x': lon, 'y': lat,
               'data': group.variables[varname][s].reshape(1, x, y),
               'time': num2date(group.variables['time'][s],
                                group.variables['time'].units)}


def get_times(time, start=None, end=None, isfile=None):
    '''Get the start and end index for a given period'''
    if type(end) == type('a') and type(start) == type('a'):
        end = datetime.strptime(end, '%Y-%m-%d %H:%M')
        start = datetime.strptime(start, '%Y-%m-%d %H:%M')
        start = date2num([start], time.units)
        end = date2num([end], time.units)
        e_idx = np.argmin(np.fabs(time[:] - end))+1
        s_idx = np.argmin(np.fabs(time[:] - start))
        return [(s_idx, e_idx)]
    elif type(isfile) == type(None):
        isifle = np.ones_like(time.shape[0])
        return spl(isfile, time)
